/*
 * *************************************************
 * Copyright (c) 2018 QMTH. All Rights Reserved.
 * Created by Deason on 2018-10-22 15:27:24.
 * *************************************************
 */

package cn.com.qmth.examcloud.core.print.service.impl;

import cn.com.qmth.examcloud.commons.base.exception.StatusException;
import cn.com.qmth.examcloud.core.print.common.jpa.SearchBuilder;
import cn.com.qmth.examcloud.core.print.common.jpa.SpecUtils;
import cn.com.qmth.examcloud.core.print.common.upyun.SystemProperty;
import cn.com.qmth.examcloud.core.print.common.upyun.UpYunClient;
import cn.com.qmth.examcloud.core.print.common.utils.*;
import cn.com.qmth.examcloud.core.print.entity.CoursePaper;
import cn.com.qmth.examcloud.core.print.entity.CourseStatistic;
import cn.com.qmth.examcloud.core.print.entity.ObjectiveQuestionStructure;
import cn.com.qmth.examcloud.core.print.entity.SubjectiveQuestionStructure;
import cn.com.qmth.examcloud.core.print.enums.PaperStatus;
import cn.com.qmth.examcloud.core.print.repository.CoursePaperRepository;
import cn.com.qmth.examcloud.core.print.repository.CourseStatisticRepository;
import cn.com.qmth.examcloud.core.print.service.CoursePaperService;
import cn.com.qmth.examcloud.core.print.service.ExamQuestionStructureService;
import cn.com.qmth.examcloud.core.print.service.bean.coursepaper.*;
import com.itextpdf.text.pdf.PdfReader;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.data.jpa.domain.Specification;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.*;

import static cn.com.qmth.examcloud.core.print.common.Constants.*;

/**
 * @author: fengdesheng
 * @since: 2018/10/17
 */
@Service
public class CoursePaperServiceImpl implements CoursePaperService {
    private static final Logger log = LoggerFactory.getLogger(CoursePaperServiceImpl.class);
    @Autowired
    private CoursePaperRepository coursePaperRepository;
    @Autowired
    private CourseStatisticRepository courseStatisticRepository;
    @Autowired
    private ExamQuestionStructureService examQuestionStructureService;
    @Autowired
    private SystemProperty systemProperty;
    @Autowired
    private UpYunClient upYunClient;

    @Value("${question.url.prefix}")
    private String QUESTION_URL_PREFIX;//题库访问地址前缀

    @Override
    public List<CoursePaperLessInfo> getCoursePaperList(CoursePaperQuery query) {
        Check.isNull(query, "查询参数不能为空！");
        Check.isNull(query.getOrgId(), "学校ID不能为空！");
        Check.isNull(query.getExamId(), "考试ID不能为空！");
        Check.isNull(query.getCourseId(), "课程ID不能为空！");

        //查询条件
        SearchBuilder searches = new SearchBuilder();
        searches.eq("orgId", query.getOrgId());
        searches.eq("examId", query.getExamId());
        if (query.getCourseId() != null) {
            searches.eq("courseId", query.getCourseId());
        }
        Specification<CoursePaper> spec = SpecUtils.buildSearchers(CoursePaper.class, searches.build());
        List<CoursePaper> list = coursePaperRepository.findAll(spec);
        return CoursePaperConvert.ofList(list);
    }

    @Override
    public void syncCourseNameByCourseId(Long courseId, String courseName) {
        if (courseId == null || StringUtils.isBlank(courseName)) {
            return;
        }
        coursePaperRepository.updateCourseNameByCourseId(courseId, courseName);
    }

    @Override
    public void syncCoursePaper(CoursePaper coursePaper) {
        Check.isNull(coursePaper.getOrgId(), "学校ID不能为空！");
        Check.isNull(coursePaper.getExamId(), "考试ID不能为空！");
        Check.isNull(coursePaper.getCourseId(), "课程ID不能为空！");
        Check.isEmpty(coursePaper.getCourseCode(), "课程代码不能为空！");
        Check.isEmpty(coursePaper.getCourseName(), "课程名称不能为空！");
        Check.isEmpty(coursePaper.getPaperId(), "试卷ID不能为空！");
        Check.isEmpty(coursePaper.getPaperName(), "试卷名称不能为空！");

        //试卷页面地址和答案页面地址
        String paperHtmlUrl = QUESTION_URL_PREFIX.concat("/api/ecs_ques/paper/pdf/").concat(coursePaper.getPaperId());
        String answerHtmlUrl = QUESTION_URL_PREFIX.concat("/api/ecs_ques/paper/answer/pdf/").concat(coursePaper.getPaperId());

        //转换PDF文件
        //final String rootDir = Constants.rootFileDir();
        final String rootDir = systemProperty.getTempDir();
        final String paperPdfPath = rootDir + "/" + FileUtils.randomUUID() + SUFFIX_PDF;
        final String answerPdfPath = rootDir + "/" + FileUtils.randomUUID() + SUFFIX_PDF;
        FileUtils.makeDirs(rootDir);
        ElectronUtils.toPdf(systemProperty.getDir(), paperHtmlUrl, paperPdfPath);
        ElectronUtils.toPdf(systemProperty.getDir(), answerHtmlUrl, answerPdfPath);

        //上传PDF至又拍云
        String paperPdfUrl = upYunClient.upload(new File(paperPdfPath));
        String answerPdfUrl = upYunClient.upload(new File(answerPdfPath));

        //获取试卷PDF的页数
        int paperP = 0;
        try (FileInputStream fis = new FileInputStream(paperPdfPath);) {
            PdfReader pdfReader = new PdfReader(fis);
            paperP = pdfReader.getNumberOfPages();
        } catch (IOException e) {
            log.error(e.getMessage());
        }
        if (paperP == 0) {
            throw new StatusException(PRT_CODE_400, "试卷页数不能为空！");
        }

        CoursePaper oldCoursePaper = coursePaperRepository.findByPaperId(coursePaper.getPaperId());
        if (oldCoursePaper != null) {
            //存在则修改
            oldCoursePaper.setPaperName(coursePaper.getPaperName());
            oldCoursePaper.setPaperHtmlUrl(paperHtmlUrl);
            oldCoursePaper.setAnswerHtmlUrl(answerHtmlUrl);
            oldCoursePaper.setPaperPdfUrl(paperPdfUrl);
            oldCoursePaper.setAnswerPdfUrl(answerPdfUrl);
            oldCoursePaper.setPaperP(paperP);
            coursePaperRepository.save(oldCoursePaper);
        } else {
            //否则新增
            coursePaper.setPaperHtmlUrl(paperHtmlUrl);
            coursePaper.setAnswerHtmlUrl(answerHtmlUrl);
            coursePaper.setPaperPdfUrl(paperPdfUrl);
            coursePaper.setAnswerPdfUrl(answerPdfUrl);
            coursePaper.setPaperP(paperP);
            coursePaperRepository.save(coursePaper);
        }

        //保存或更新试卷试题结构
        examQuestionStructureService.savePaperQuestionStructure(coursePaper.getExamId(), coursePaper.getPaperId());

        //如果当前考试课程"只有一种试卷类型"且"未分配试卷"，则默认分配该试卷
        SearchBuilder searches = new SearchBuilder()
                .eq("orgId", coursePaper.getOrgId())
                .eq("examId", coursePaper.getExamId())
                .eq("courseId", coursePaper.getCourseId());
        Specification<CourseStatistic> spec = SpecUtils.buildSearchers(CourseStatistic.class, searches.build());
        List<CourseStatistic> statistics = courseStatisticRepository.findAll(spec);
        if (statistics == null || statistics.isEmpty()) {
            return;
        }

        for (CourseStatistic statistic : statistics) {
            if (PaperStatus.已有.getIndex() == statistic.getPaperStatus()) {
                //跳过已分配的
                continue;
            }
            if (statistics.size() == 1) {
                //只有一种试卷类型，则默认分配
                statistic.setCoursePaper(coursePaper);
                statistic.setPaperStatus(PaperStatus.已有.getIndex());
                break;
            }
            //否则，分配状态改为"未指定"
            statistic.setPaperStatus(PaperStatus.未指定.getIndex());
        }
        courseStatisticRepository.save(statistics);
    }

    @Transactional
    public void allotCoursePaper(Long courseStatisticId, Long coursePaperId) {
        Check.isEmpty(courseStatisticId, "课程统计ID不能为空！");
        Check.isEmpty(coursePaperId, "课程试卷ID不能为空！");

        CourseStatistic statistic = courseStatisticRepository.findOne(courseStatisticId);
        Check.isNull(statistic, "当前课程统计不存在！");

        CoursePaper coursePaper = coursePaperRepository.findOne(coursePaperId);
        Check.isNull(coursePaper, "当前课程试卷不存在！");

        statistic.setCoursePaper(coursePaper);
        statistic.setPaperStatus(PaperStatus.已有.getIndex());
        courseStatisticRepository.save(statistic);
    }

    /**
     * (整体)分配待指定试卷
     * 选定学校和考试后，为该考试下的所有待分配课程随机分配试卷
     * 当题库传输过来的试卷数量不足时，会提示试卷数不足的课程名称(代码)
     * 当题库传输过来的试卷数量≥需求时，每种试卷类型随机分配
     * 随机分配后的“课程代码 + 类型”其试卷状态变为"已有"
     * 试卷状态：已有、无、未指定三种，显示为“未指定”时，点击时弹出试卷制定框
     */
    @Override
    public void allotAllCoursePaper(Long orgId, Long examId) {
        //查询所有未指定试卷的考试课程
        SearchBuilder searches = new SearchBuilder()
                .eq("orgId", orgId)
                .eq("examId", examId)
                .eq("paperStatus", PaperStatus.未指定.getIndex());

        Specification<CourseStatistic> spec = SpecUtils.buildSearchers(CourseStatistic.class, searches.build());
        List<CourseStatistic> statistics = courseStatisticRepository.findAll(spec);
        if (statistics == null || statistics.isEmpty()) {
            log.warn("No course paper need to allot.");
            return;
        }

        //反馈信息
        StringBuilder errors = new StringBuilder();

        //获取考试课程下的试卷列表
        List<CoursePaper> coursePapers = this.getCoursePapers(orgId, examId);

        //按课程封装试卷列表 Map<courseId, List<CoursePaper>>
        Map<Long, List<CoursePaper>> maps = new HashMap<>();
        for (CoursePaper coursePaper : coursePapers) {
            List<CoursePaper> papers = maps.get(coursePaper.getCourseId());
            if (papers == null) {
                papers = new ArrayList<>();
                maps.put(coursePaper.getCourseId(), papers);
            }
            papers.add(coursePaper);
        }

        //按课程随机指定试卷
        List<CourseStatistic> needUpdateStatistics = new ArrayList<>();
        for (CourseStatistic statistic : statistics) {
            List<CoursePaper> papers = maps.get(statistic.getCourseId());
            if (papers == null || papers.isEmpty()) {
                errors.append(String.format("《%s(%s)》", statistic.getCourseName(), statistic.getCourseCode()));
                continue;
            }
            Collections.shuffle(papers);
            statistic.setCoursePaper(papers.get(0));
            statistic.setPaperStatus(PaperStatus.已有.getIndex());
            needUpdateStatistics.add(statistic);
        }

        //批量更新
        if (!needUpdateStatistics.isEmpty()) {
            courseStatisticRepository.save(needUpdateStatistics);
        }

        //试卷不足时，反馈给前端
        if (errors.length() > 0) {
            errors.append("等课程的试卷数不足！");
            throw new StatusException(PRT_CODE_500, errors.toString());
        }
    }

    private List<CoursePaper> getCoursePapers(Long orgId, Long examId) {
        SearchBuilder searches = new SearchBuilder();
        searches.eq("orgId", orgId);
        searches.eq("examId", examId);
        Specification<CoursePaper> spec = SpecUtils.buildSearchers(CoursePaper.class, searches.build());
        return coursePaperRepository.findAll(spec);
    }

    @Override
    public CoursePaperTotalInfo getPaperTotalByOrgIdAndExamId(Long orgId, Long examId) {
        Check.isNull(orgId, "学校ID不能为空！");
        Check.isNull(examId, "考试ID不能为空！");

        //试卷总数 = 试卷数据中"课程代码 + 试卷类型"的数量
        int allNum = courseStatisticRepository.countByOrgIdAndExamId(orgId, examId);

        //已有试卷数 = 已分配待指定试卷的数量
        int existNum = courseStatisticRepository.countByOrgIdAndExamIdAndPaperStatus(orgId, examId, PaperStatus.已有.getIndex());

        //缺少试卷数 = 试卷总数 - 已有试卷数
        int missNum = allNum - existNum;

        return new CoursePaperTotalInfo(allNum, existNum, missNum);
    }

    @Override
    public File exportBatchCoursePaper(ExportBatchReq req) {
        Check.isNull(req, "参数不能为空！");
        Check.isEmpty(req.getIds(), "至少选择一门考试课程！");
        Check.isFalse(req.required(), "至少选择一种导出内容！");

        SearchBuilder searches = new SearchBuilder();
        searches.eq("paperStatus", PaperStatus.已有.getIndex());
        if (req.getIds().size() == 1) {
            searches.eq("id", req.getIds().get(0));
        } else {
            searches.in("id", req.getIds());
        }
        Specification<CourseStatistic> spec = SpecUtils.buildSearchers(CourseStatistic.class, searches.build());

        List<CourseStatistic> statistics = courseStatisticRepository.findAll(spec);
        return this.doExport(statistics, req.getNeedPaper(), req.getNeedAnswer(), req.getNeedStruct());
    }

    @Override
    public File exportAllCoursePaper(ExportAllReq req) {
        Check.isNull(req, "参数不能为空！");
        Check.isNull(req.getOrgId(), "学校ID不能为空！");
        Check.isNull(req.getExamId(), "考试ID不能为空！");
        Check.isFalse(req.required(), "至少选择一种导出内容！");

        SearchBuilder searches = new SearchBuilder();
        searches.eq("orgId", req.getOrgId());
        searches.eq("examId", req.getExamId());
        searches.eq("paperStatus", PaperStatus.已有.getIndex());
        Specification<CourseStatistic> spec = SpecUtils.buildSearchers(CourseStatistic.class, searches.build());

        List<CourseStatistic> statistics = courseStatisticRepository.findAll(spec);
        return this.doExport(statistics, req.getNeedPaper(), req.getNeedAnswer(), req.getNeedStruct());
    }

    private File doExport(List<CourseStatistic> statistics, boolean needPaper, boolean needAnswer, boolean needStruct) {
        if (statistics == null || statistics.isEmpty()) {
            log.warn("No paper to export!");
            return null;
        }

        //封装待导出的试卷或答案文件信息列表
        List<ExportFileInfo> exportFiles = new ArrayList<>();

        //封装试卷结构-客观题数据
        List<ObjectiveQuestionStructure> allObjectives = new ArrayList<>();

        //封装试卷结构-主观题数据
        List<SubjectiveQuestionStructure> allSubjectives = new ArrayList<>();

        for (CourseStatistic statistic : statistics) {
            CoursePaper paper = statistic.getCoursePaper();
            if (paper == null) {
                continue;
            }

            //命名规则：试卷名_课程代码_试卷类型_试卷或答案
            final String title = String.format("%s_%s_%s_", paper.getPaperName(), paper.getCourseCode(), statistic.getPaperType());
            ExportFileInfo info = new ExportFileInfo();

            if (needPaper) {
                info.setPaperPdf(new Pair<>(title + PAPER_PDF_NAME, paper.getPaperPdfUrl()));
            }

            if (needAnswer) {
                info.setAnswerPdf(new Pair<>(title + ANSWER_PDF_NAME, paper.getAnswerPdfUrl()));
            }

            if (needStruct) {
                //客观题结构
                List<ObjectiveQuestionStructure> objectives = examQuestionStructureService.getObjectiveQuestionStructureList(statistic.getOrgId(), statistic.getExamId(), paper.getPaperId(), statistic.getPaperType());
                allObjectives.addAll(objectives);

                //主观题结构
                List<SubjectiveQuestionStructure> subjectives = examQuestionStructureService.getSubjectiveQuestionStructureList(statistic.getOrgId(), statistic.getExamId(), paper.getPaperId(), statistic.getPaperType());
                allSubjectives.addAll(subjectives);
            }

            exportFiles.add(info);
        }

        //文件存放目录
        //final String targetDir = Constants.rootFileDir() + "/" + FileUtils.randomUUID();
        final String targetDir = systemProperty.getTempDir() + "/" + FileUtils.randomUUID();
        FileUtils.makeDirs(targetDir);

        //优先处理试卷或答案文件(需要网络下载)
        if (needPaper || needAnswer) {
            this.doPaperAndAnswerFiles(exportFiles, targetDir);
        }

        //处理试卷结构文件
        if (needStruct) {
            this.doStructureFiles(allObjectives, allSubjectives, targetDir);
        }

        log.debug("All files finished.");

        //打包所有文件
        return packageFiles(targetDir);
    }

    private void doPaperAndAnswerFiles(List<ExportFileInfo> exportFiles, final String targetDir) {
        if (exportFiles.isEmpty()) {
            return;
        }

        //试卷或答案存放目录
        final String tempDir = targetDir + "/paper";
        FileUtils.makeDirs(tempDir);

        //下载试卷或答案文件
        for (ExportFileInfo info : exportFiles) {
            this.loadFile(tempDir, info.getPaperPdf());
            this.loadFile(tempDir, info.getAnswerPdf());
        }
    }

    private void doStructureFiles(List<ObjectiveQuestionStructure> allObjectives, List<SubjectiveQuestionStructure> allSubjectives, final String targetDir) {
        final String objectiveFile = targetDir + "/" + OBJECTIVE_EXCEL_NAME;
        ExcelUtils.exportExcel(ObjectiveQuestionStructure.class, allObjectives, new File(objectiveFile), OBJECTIVE_TITLE);

        final String subjectiveFile = targetDir + "/" + SUBJECTIVE_EXCEL_NAME;
        ExcelUtils.exportExcel(SubjectiveQuestionStructure.class, allSubjectives, new File(subjectiveFile), SUBJECTIVE_TITLE);
    }

    private File packageFiles(final String targetDir) {
        //压缩打包所有文件
        File target = new File(targetDir);
        File zipFile = new File(targetDir + SUFFIX_ZIP);
        FileUtils.zip(target, zipFile);

        try {
            if (!zipFile.exists()) {
                zipFile.createNewFile();
            }
            //压缩完后移除临时文件
            org.apache.commons.io.FileUtils.deleteDirectory(target);
        } catch (IOException e) {
            log.error(e.getMessage());
        }
        return zipFile;
    }

    private void loadFile(String dir, Pair<String, String> pair) {
        if (pair == null) {
            //未指定时则不用下载
            return;
        }
        //key为文件名,value为文件链接地址
        final String filePath = dir + "/" + pair.getKey();
        FileUtils.saveURLToFile(pair.getValue(), new File(filePath));
    }

}