/*
 * *************************************************
 * Copyright (c) 2018 QMTH. All Rights Reserved.
 * Created by Deason on 2018-10-17 17:16:01.
 * *************************************************
 */

package cn.com.qmth.examcloud.core.print.test;

import cn.com.qmth.examcloud.core.print.common.jpa.OrderBuilder;
import cn.com.qmth.examcloud.core.print.common.jpa.SearchBuilder;
import cn.com.qmth.examcloud.core.print.common.jpa.SpecUtils;
import cn.com.qmth.examcloud.core.print.entity.PrintingProject;
import cn.com.qmth.examcloud.core.print.repository.PrintingProjectRepository;
import cn.com.qmth.examcloud.core.print.service.PrintingProjectService;
import cn.com.qmth.examcloud.core.print.service.bean.common.ExamInfo;
import cn.com.qmth.examcloud.core.print.service.bean.common.OrgInfo;
import cn.com.qmth.examcloud.core.print.service.bean.printingproject.PrintingProjectInfo;
import cn.com.qmth.examcloud.core.print.service.bean.printingproject.PrintingProjectQuery;
import org.junit.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.Page;
import org.springframework.data.jpa.domain.Specification;

import java.util.Date;
import java.util.List;

/**
 * @author: fengdesheng
 * @since: 2018/10/17
 */
public class PrintingProjectServiceTest extends BaseTest {
    @Autowired
    private PrintingProjectRepository printingProjectRepository;
    @Autowired
    private PrintingProjectService printingProjectService;

    @Test
    public void getPrintingProjectListTest() throws Exception {
        PrintingProjectQuery query = new PrintingProjectQuery();
        query.setOrgId(orgId);
        query.setExamId(examId);
        query.setPmId(1L);
        query.setSupplierId(1L);
        Page<PrintingProjectInfo> page = printingProjectService.getPrintingProjectList(query);
        System.out.println(jsonMapper.toJson(page));
    }

    @Test
    public void updatePrintingProjectTest() throws Exception {
        PrintingProjectInfo info = new PrintingProjectInfo();
        info.setId(orgId);
        info.setPmId(examId);
        info.setPmName("雷布斯");
        info.setSupplierId(1L);
        info.setSupplierName("小米公司");
        info.setPrepareStartTime(new Date());
        info.setPrepareEndTime(new Date());
        info.setPrintStartTime(new Date());
        info.setPrintEndTime(new Date());
        info.setMailStartTime(new Date());
        info.setMailEndTime(new Date());
        printingProjectService.updatePrintingProject(info);
    }

    @Test
    public void getOrgExamTest() throws Exception {
        Long pmId = null, supplierId = null;
        List<OrgInfo> orgList = printingProjectService.getOrgList(pmId, supplierId);
        System.out.println(jsonMapper.toJson(orgList));

        Long orgId = null;
        List<ExamInfo> examList = printingProjectService.getExamList(orgId);
        System.out.println(jsonMapper.toJson(examList));
    }

    @Test
    public void syncPrintingProjectTest() throws Exception {
        ExamInfo examInfo = new ExamInfo(orgId, "武汉大学", examId, "计算机考试");
        printingProjectService.syncPrintingProject(examInfo);
        printingProjectService.syncOrgNameByOrgId(orgId, "武汉大学1");
        printingProjectService.syncExamNameByExamId(examId, "计算机考试1");
        printingProjectService.syncPmNameByPmId(1L, "雷布斯1");
        printingProjectService.syncSupplierNameBySupplierId(1L, "小米公司1");
    }

    //@Test
    public void specTest() throws Exception {
        //查询条件1
        SearchBuilder searchBuilder1 = new SearchBuilder().eq("orgName", "武汉大学");
        Specification<PrintingProject> spec1 = SpecUtils.buildSearchers(PrintingProject.class, searchBuilder1.build());

        //查询条件2
        SearchBuilder searchBuilder2 = new SearchBuilder().eq("orgId", orgId).eq("examId", examId);
        Specification<PrintingProject> spec2 = SpecUtils.buildSearchers(PrintingProject.class, searchBuilder2.build(), true);

        //查询条件3
        SearchBuilder searchBuilder3 = new SearchBuilder().eq("pmId", 1L).eq("supplierId", 1L);
        Specification<PrintingProject> spec3 = SpecUtils.buildSearchers(PrintingProject.class, searchBuilder3.build());

        //合并查询条件
//        Specification<PrintingProject> spec = SpecUtils.orMerge(spec1, spec2, spec3);
//        Specification<PrintingProject> spec = SpecUtils.andMerge(spec1, spec2, spec3);
        Specification<PrintingProject> spec = SearchBuilder.EQ("orgId", orgId);

        OrderBuilder orderBuilder = new OrderBuilder().desc("id");
        List<PrintingProject> list = printingProjectRepository.findAll(spec, orderBuilder.build());
        System.out.println(jsonMapper.toJson(list));
    }

}